
module eshop

--------------------------------------------------

sig Book {}

sig BookDB {
  available:   set Book,
--backordered: set Book,
  purchased:   set Book
}

-- there is only 1 book database
fact {
  one BookDB
}

-- there are book(s) in database
fact {
  all db:BookDB | some book:Book | {
    book in (db.available + db.purchased)
  }
}

fact {
  all book:Book | some db:BookDB | {
    book in (db.available + db.purchased)
  }
}

fact {
  all db:BookDB, book:Book | {
    ! book in (db.purchased & db.available)
  } 
}
 
fun initDB(db:BookDB) {
    all book:Book | ! book in db.purchased
    some book:Book | book in db.available
}

--------------------------------------------------

sig ShopperID {}

sig Shopper {
  id:          ShopperID,
  detailed:    Book,
  listed:      set Book,
  incart:      set Book,
--backordered: set Book,
  purchased:   set Book
}

fact {
  some Shopper
}

fact {
  all shopper:Shopper, book:Book |
    book = shopper.detailed => book in shopper.listed
}

fact {
  all shopper:Shopper, book:Book |
    ! book in (shopper.incart & shopper.purchased)
}

fact {
  all shopper1,shopper2:Shopper, book:Book | {
    shopper1 != shopper2 =>
      ! book in (shopper1.purchased & shopper2.purchased)
  }
}

fact {
  all db:BookDB, book:Book |
    book in db.purchased iff
      some shopper:Shopper | book in shopper.purchased
}

fun initShopper(shopper:Shopper) {
  all book:Book |
    ! book in shopper.detailed
    && ! book in shopper.listed
    && ! book in shopper.incart
--  && ! book in shopper.backordered
    && ! book in shopper.purchased
}

fun emptyCart(shopper:Shopper) {
  no shopper.incart
}

fun addToCart(shopper:Shopper, db:BookDB, book:Book) : Shopper {
  book in db.available => 
    shopper.incart = shopper.incart + book
  result = shopper
}

fun placeOrder(shopper:Shopper) : Shopper {
  all book:Book | some db:BookDB | {
      book in (db.available & shopper.incart) =>
        db.available = db.available - book
        shopper.incart = shopper.incart - book
	shopper.purchased = shopper.purchased + book
  }
  result = shopper
}

--------------------------------------------------

--sig Content {}

--------------------------------------------------

sig Operation {}

sig NoOp extends Operation {}

sig EmptyCart extends Operation {}

sig AddToCart extends Operation {}

sig PlaceOrder extends Operation {}

--------------------------------------------------

sig Page {
--content: Content,
  links:   set Link
}

sig Link {
  operation:   Operation,
  destination: Page
}

sig InitialPage extends Page {}

sig Browsing extends Page {}
sig ListView extends Browsing {}
sig DetailView extends Browsing {}

sig OrderControl extends Page {}
sig CartView extends OrderControl {}
sig Confirm extends OrderControl {}

fact {
  some Page
}

fact {
     no (InitialPage & Browsing)
  && no (InitialPage & OrderControl)
  && no (Browsing & OrderControl)
  && no (ListView & DetailView)
  && no (CartView & Confirm)
}

--------------------------------------------------

sig EShop {
  pages:     set Page,
  firstpage: InitialPage,
  currpage:  Page,
  shopper:   Shopper
}

fact {
  some EShop
}

--fact {
--  all shopper:Shopper | some eshop:EShop | shopper = eshop.shopper
--}

fact {
  all page:Page | some eshop:EShop | page in eshop.pages
}

fact {
  all eshop:EShop | eshop.firstpage in eshop.pages
}

fact {
  all eshop:EShop | eshop.currpage in eshop.pages
}

fact {
  all eshop:EShop | 
    some p1,p2,p3,p4,p5:eshop.pages | {
      p1 in InitialPage 
      p2 in ListView
      p3 in DetailView
      p4 in CartView
      p5 in Confirm
    }
}

fact {
  all eshop:EShop | 
    some p:eshop.pages, l1,l2:p.links, op1,op2:Operation, p1,p2:eshop.pages | {
      l1.operation   = op1
      l1.destination = p1
      l2.operation   = op2
      l2.destination = p2

      (p in InitialPage &&    (op1 = NoOp       && p1 = ListView))
    }
}

fact {
  all eshop:EShop | 
    some p:eshop.pages, l1,l2:p.links, op1,op2:Operation, p1,p2:eshop.pages | {
      l1.operation   = op1
      l1.destination = p1
      l2.operation   = op2
      l2.destination = p2

      (p in ListView    &&    (op1 = NoOp       && p1 = DetailView
                            && op2 = AddToCart  && p2 = CartView))
    }
}

fact {
  all eshop:EShop | 
    some p:eshop.pages, l1,l2:p.links, op1,op2:Operation, p1,p2:eshop.pages | {
      l1.operation   = op1
      l1.destination = p1
      l2.operation   = op2
      l2.destination = p2

      (p in DetailView  &&    (op1 = NoOp       && p1 = ListView
                            && op2 = AddToCart  && p2 = CartView))
    }
}

fact {
  all eshop:EShop | 
    some p:eshop.pages, l1,l2:p.links, op1,op2:Operation, p1,p2:eshop.pages | {
      l1.operation   = op1
      l1.destination = p1
      l2.operation   = op2
      l2.destination = p2

      (p in CartView    &&    (op1 = NoOp       && p1 = Confirm
                            && op2 = NoOp       && p2 = Browsing))
    }
}

fact {
  all eshop:EShop | 
    some p:eshop.pages, l1,l2:p.links, op1,op2:Operation, p1,p2:eshop.pages | {
      l1.operation   = op1
      l1.destination = p1
      l2.operation   = op2
      l2.destination = p2

      (p in Confirm     &&    (op1 = PlaceOrder && p1 = InitialPage
                            && op2 = NoOp       && p2 = ListView))
    }
}

/*
fact {
  all eshop:EShop, p:eshop.pages, link:p.links | 
    some op:Operation, p':eshop.pages | {
      link.operation   = op
      link.destination = p'

    (p in InitialPage => p' in ListView)
    (p in ListView    => p' in (DetailView + CartView))
    (p in DetailView  => p' in (ListView   + CartView))
    (p in CartView    => p' in (Browsing   + Confirm))
    (p in Confirm     => p' in (ListView   + InitialPage))

    (op in NoOp 
       iff  (p in InitialPage && p' in ListView)
	 || (p in ListView    && p' in DetailView)
	 || (p in DetailView  && p' in ListView)
	 || (p in CartView    && p' in Browsing)
	 || (p in CartView    && p' in Confirm)
	 || (p in Confirm     && p' in ListView))
    (op in AddToCart 
       iff  (p in Browsing    && p' in CartView))
    (op in PlaceOrder 
       iff  (p in Confirm     && p' in InitialPage))
    }
}
*/

fun initEShops() {
}

--------------------------------------------------

sig Tick {}

sig EShopTrace {
  ticks: set Tick,
  first: ticks,
  last:  ticks,
  next:  (ticks-last)! -> !(ticks-first),
  state: ticks -> !EShop
}
{
all trace:EShopTrace | {
  
some db:BookDB | { 
initDB(db)
all shopper:Shopper | 
  initShopper(shopper)
  initEShops()

  first.*next = ticks

  (all t:ticks-last | 
    some e,e':EShop, p,p':Page, link:Link, op:Operation, book:Book | {
    e = t.state && e' = t.next.state 

    e.pages = e'.pages
    e.firstpage = e'.firstpage
    e.shopper.id = e'.shopper.id

    e.currpage = p && e'.currpage = p'
    link in p.links && link.operation = op && link.destination = p'

    op in AddToCart => addToCart(e.shopper, db, book) = e'.shopper
    op in PlaceOrder => placeOrder(e.shopper) = e'.shopper
    })
  }
}
}

--------------------------------------------------
--------------------------------------------------
--------------------------------------------------

assert BookInOneCart {
  some book:Book, s1,s2:Shopper |
    s1 != s2 && book in (s1.incart & s2.incart)
}
check BookInOneCart for 2 but 5 Page, 10 Link, 10 Operation

--------------------------------------------------

assert BookHasMultipleOwners {
  some db:BookDB | all s1,s1',s1'',s2,s2',s2'':Shopper | {
    s1 != s2 => some book:Book | {

      s1'  = addToCart(s1, db, book)
      s2'  = addToCart(s2, db, book)
    
      s1'' = placeOrder(s1')
      s2'' = placeOrder(s2')

      book in (s1''.purchased & s2''.purchased)
    }
  }
}
check BookHasMultipleOwners for 2 but 5 Page, 10 Link, 10 Operation

--------------------------------------------------

assert BookDBDoesntMatchShoppers {
  some db:BookDB | 
    # {book:Book | book in db.purchased}
    != # {book:Book | some shopper:Shopper | book in shopper.purchased}
}
check BookDBDoesntMatchShoppers for 2 but 5 Page, 2 Link, 2 Operation

--------------------------------------------------


